unit Dial02;

(*  This software is FREEWARE

  Delphi SNTP Client
  -----------------
  Developed on :  D6.00 Enterprise
  Tested on    :  WIN-NT4/SP6, WIN2K, WIN98se

  ================================================================
                    This software is FREEWARE
                    -------------------------
  If this software works, it was surely written by Dirk Claessens
                   <dirkcl@pandora.be>
  (If it does'nt, I don't know anything about it. :-)
  ================================================================

  Installation notes
  ==================
  1. Please, FIRST install Francois Piette's WSocket component
     BEFORE you compile this unit.
     Component> Install Component > browse to wsocket.pas, etc...
  2. All SNTP-servers included were checked to be operational at the time of
     writing, and offer _open_ access.
     If you modify the list, please note that not all SNTP servers
     have open access. Also, some servers may limit the number of requests
     on an hourly or daily basis, so do _NOT_ flood a server with requests.
  3. The calculated clock correction takes propagation delay into account.
     However, symmetric request/reply delay is assumed. Therefore, choose
     a server near to you for highest accuracy.
  4. If you have D3/D4/D5, you probably will have compilation problems, due to
     some new form properties in D6. Try to "ignore" the errors.

*)



interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, Buttons, StdCtrls,  WinSock, DateUtils, ExtCtrls, Menus,
  OverbyteIcsWndControl, OverbyteIcsWSocket;


const
   Version   = '1.1';
   SNTP_PORT = '123';  // standard SNTP-port RFC 2030
   ONE_SEC   = 1/SecsPerDay;
type

  TNTPTimeStamp = packed record // SNTP Time stamp format,  see RFC 1305
    Secs: DWORD;                // seconds since  January 1, 1900, 00:00
    Frac: DWORD;                // fraction of second
  end;

  TNtpRec = packed record      // SNTP record, see RFC 2030
       LiVuMode        : byte;
       Stratum,                // stratum, 1..3
       Poll,                   // not used
       Precision       : byte; //  -"-
       RootDelay,              //  -"-
       RootDispersion,         //  -"-
       RefID           : array[0..3] of char;   // -"-
       RefTStamp,                        // time server was last synced
       OrgTStamp,                        // originate timestamp of client
       RxTStamp,                         // time server received client request
       TxTStamp        : TNTPTimeStamp;  // time server sent reply
  end;



  TWinTStamp = record
     DayTime : TDateTime;
     MSecs   : integer;
  end;



  TDlg02 = class(TForm)
    cbServers: TComboBox;
    StaticText1: TStaticText;
    edtLocalTime: TEdit;
    edtError: TEdit;
    Timer1: TTimer;
    StaticText2: TStaticText;
    StaticText3: TStaticText;
    edtRTT: TEdit;
    StaticText4: TStaticText;
    cbInterval: TComboBox;
    StaticText5: TStaticText;
    StaticText6: TStaticText;
    cbSync: TCheckBox;
    stStatus: TStaticText;
    EditTime1: TEdit;
    StaticText7: TStaticText;
    UDPSock: TWSocket;
    BitBtn1: TBitBtn;
    EditTime2: TEdit;
    procedure UDPSockDataAvailable(Sender: TObject; Error: Word);
    procedure UDPSockSessionConnected(Sender: TObject; Error: Word);
    procedure Timer1Timer(Sender: TObject);
    procedure UDPSockError(Sender: TObject);
    procedure cbServersSelect(Sender: TObject);
    procedure cbIntervalSelect(Sender: TObject);
    procedure BitBtn1Click(Sender: TObject);
    procedure FormCreate(Sender: TObject);
  private
    { Private declarations }
    CurrentServer : string;
    ServerTime    : TDateTime;
    SendRec : TNTPRec;
    T1      : TNTPTimeStamp;
    IpResolved : boolean;
    SyncTicks  : integer;
    procedure ReverseByteOrder( var NtpTime: DWORD );
    function  Ntp2Double( NtpTime: TNTPTimestamp ): Double;
    function  NowToNTP : TNtpTimeStamp;
    procedure ResolveIP;
    procedure RequestTime;
  public
    { Public declarations }
  end;

var
  Dlg02: TDlg02;

implementation

{$R *.dfm}

procedure TDlg02.FormCreate(Sender: TObject);
begin
  SyncTicks := 5;
  cbServers.ItemIndex := 0;
  cbSync.Checked := true;
  cbServersSelect(Sender);
  Timer1Timer (Sender);
  RequestTime;
  Timer1Timer (Sender);
end;

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *//
// swap timestamp  from network byte order to Intel byte order
procedure TDlg02.ReverseByteOrder( var NtpTime: DWORD );
var
   Ta: array[0..3] of byte absolute NTPTime;
   Temp : WORD;
begin
  temp  := ta[0];     // 1 2 3 4 => 4 3 2 1
  ta[0] := ta[3];
  ta[3] := temp;
  Temp  := ta[1];
  ta[1] := ta[2];
  ta[2] := temp;
end;



//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *//
// calculates Delphi float from NTP timestamp
// integer part = seconds since jan 1, 1900 00:00 Hr
// fract. part  = fraction of second
function  TDlg02.Ntp2Double( NtpTime: TNTPTimestamp ): Double;
begin
  ReverseByteOrder( NTPTime.Secs );
  ReverseByteOrder( NTPTime.Frac );
  Result  := NTpTime.Secs;
  Result := Result +  ( NTPTime.Frac/$FFFFFFFF) ;
end;

procedure TDlg02.Timer1Timer(Sender: TObject);
begin
   Exit; // provisoire bug2012

   
   //
   edtLocalTime.Text := DateTimeTostr(Now);
   //
   if ServerTime <> 0.0 then
   begin
     ServerTime := ServerTime + ONE_SEC;
     EditTime1.Text := DateTimeToStr(ServerTime);
     EditTime2.Text := EditTime1.Text;
   end;
   //
   if ( CurrentServer <> '')
   and not IPResolved then
      ResolveIP;    // get IP-address of server
   //
   if IpResolved             // only dec. ticks when IP resolved...
   and ( SyncTicks > 0 )then
   begin
    dec(SyncTicks);
    stStatus.Caption := IntToStr(SyncTicks) + ' secs. to next request...';
    //
    if SyncTicks = 0 then // time to do some work ?
    begin
      RequestTime;
      SyncTicks := StrToInt( cbInterval.Items[cbInterval.ItemIndex]);
    end;
   end;

end;



//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *//
procedure TDlg02.RequestTime;
begin
  stStatus.Caption := 'requesting time...';
  FillChar( SendRec, SizeOf(SendRec), #0 ); // zero out record

  // this field _must_ be set correctly , or server will not reply!
  // note: SNTP version 4 servers will also honour version 3 requests,
  // but not vice-versa
  SendRec.LiVuMode := $33; // NTP Version = 3; Mode = 3 (client)
  T1 := NowToNTP;  // Note time our request was sent
  UDPSock.Send( @SendRec, SizeOf(SendRec));
end;


//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *//
procedure TDlg02.UDPSockDataAvailable(Sender: TObject; Error: Word);
var
  Src     : TSockAddrIn;
  Len     : Integer;
  SrcLen  : Integer;
  T4      : TNtpTimeStamp;
  NTPRec  : TNTPRec;
  T1f,T2f,T3f,t4f,
  Rtt, ClockErr : Double;
  CorrectedTime : TSystemTime;
begin

  T4 := NowToNtp; // note reception time
  //
  stStatus.Caption := 'time received.';
  SrcLen := SizeOf( Src );
  Len := UDPSock.ReceiveFrom( @NTPRec, SizeOf(NtpRec), Src, SrcLen );
  //
  T1f := Ntp2Double(T1);
  T2f := Ntp2Double(NtpRec.RxTStamp);
  T3f := Ntp2Double(NtpRec.TxTStamp);
  T4f := Ntp2Double(T4);

  // calculate round trip time
  Rtt := (T4f-T1f) - (T2f-T3f);

  // calculate local clock error
  ClockErr := (( T2f - T1f ) + (T3f-T4f)) / 2;
  edtError.Text := Format('%8.3f', [-ClockErr]);
  edtRTT.Text := IntToStr(Round(Rtt*1000));
  //
   with CorrectedTime do
        DecodeDateTime( Now + (ClockErr/SecsPerDay), wYear, wMonth, wDay,
                        wHour, wMinute, wSecond, wMilliseconds );
  ServerTime := Now + (ClockErr/SecsPerDay);

  //   sync clock ?
  if  cbSync.Checked then
  begin
    SetLocalTime( CorrectedTime );
    stStatus.Caption := 'clock corrected.';
    //cbSync.Checked := false;  // avoid endless loop...
    RequestTime; // call again to see result immediately.
  end;

end;


//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *//
procedure TDlg02.UDPSockSessionConnected(Sender: TObject; Error: Word);
begin
  stStatus.Caption := UDPSock.Addr + ' resolved.';
  IpResolved := true;
  RequestTime;
end;

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //
// calculates NTP timestamp from local clock
function  TDlg02.NowToNTP : TNtpTimeStamp;
var
  T   : TSystemTime;
  TNow: TDateTime;
begin
 //
 GetSystemTime( T );
 TNow := EncodeDateTime(T.wYear, T.wMonth, T.wDay, T.wHour, T.wMinute, T.wSecond, 0);

 // Delphi's TDateTime start date = dec 30, 1899;
 // NTP uses jan 1, 1900 00:00 !
 Result.Secs := Round((TNow - 2 ) *  SecsPerDay);
 //
 Result.Frac := Round( $FFFFFFFF * (T.wMilliseconds/1000));
 ReverseByteOrder( Result.Secs );
 ReverseByteOrder( Result.Frac );
end;


//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *//
procedure TDlg02.UDPSockError(Sender: TObject);
begin
 stStatus.Caption :='error '+ IntToStr(WSAGetLastError);
 cbServers.ItemIndex := -1;
 CurrentServer := '';
 IpResolved := false;
 cbInterval.ItemIndex := 0;
end;


//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *//
procedure TDlg02.ResolveIP;
begin
  stStatus.Caption := 'resolving ' + cbServers.Items[cbservers.itemindex] + ' ...';
  UDPSock.Close;
  UDPSock.Proto := 'udp';
  UDPSock.Port := SNTP_PORT;
  UDPSock.Addr := CurrentServer;
  UDPSock.Connect;
end;

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *//
procedure TDlg02.cbServersSelect(Sender: TObject);
begin
  IpResolved := false;
  CurrentServer := cbServers.Items[cbServers.ItemIndex];
  if cbInterval.ItemIndex > 0 then
     SyncTicks := StrToInt( cbInterval.Items[cbInterval.ItemIndex]);
end;

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *//
procedure TDlg02.cbIntervalSelect(Sender: TObject);
begin
  Timer1.Enabled := false;
  SyncTicks := StrToInt( cbInterval.Items[cbInterval.ItemIndex]);
  Timer1.Interval := 1000;
  Timer1.Enabled  := true;
  if IPResolved then
      RequestTime;
end;



procedure TDlg02.BitBtn1Click(Sender: TObject);
begin
  Close;
end;

end.
